@@ -1,6 +1,6 @@
 /*******************************************************************
  * JPEGoptim
- * Copyright (c) Timo Kokkonen, 1996-2022.
+ * Copyright (c) Timo Kokkonen, 1996-2023.
  * All Rights Reserved.
  *
  * requires libjpeg (Independent JPEG Group's JPEG software
@@ -60,14 +60,15 @@
 #include <time.h>
 #include <math.h>
 
+#include "jpegmarker.h"
 #include "jpegoptim.h"
 
 
-#define VERSIO "1.5.1"
+#define VERSION "1.5.4"
 #define COPYRIGHT  "Copyright (C) 1996-2023, Timo Kokkonen"
 
 #if HAVE_WAIT && HAVE_FORK
-#define PARALLEL_PROCESSING
+#define PARALLEL_PROCESSING 1
 #define MAX_WORKERS 256
 #endif
 
@@ -79,8 +80,6 @@
 		buf=NULL;				\
 	}
 
-#define STRNCPY(dest,src,n) { strncpy(dest,src,n); dest[n-1]=0; }
-
 struct my_error_mgr {
 	struct jpeg_error_mgr pub;
 	jmp_buf setjmp_buffer;
@@ -88,7 +87,8 @@ struct my_error_mgr {
 };
 typedef struct my_error_mgr * my_error_ptr;
 
-#ifdef HAVE_FORK
+
+#ifdef PARALLEL_PROCESSING
 struct worker {
 	pid_t pid;
 	int   read_pipe;
@@ -129,12 +129,14 @@ int arith_mode = -1;
 #endif
 int max_workers = 1;
 int nofix_mode = 0;
+int files_stdin = 0;
+FILE *files_from = NULL;
 
 int compress_err_count = 0;
 int decompress_err_count = 0;
 int global_error_counter = 0;
 char last_error[JMSG_LENGTH_MAX+1];
-FILE* jpeg_log_fh;
+FILE *jpeg_log_fh;
 long average_count = 0;
 double average_rate = 0.0;
 double total_save = 0.0;
@@ -175,6 +177,8 @@ struct option long_options[] = {
 	{"size",1,0,'S'},
 	{"stdout",0,&stdout_mode,1},
 	{"stdin",0,&stdin_mode,1},
+	{"files-stdin",0,&files_stdin,1},
+	{"files-from",1,0,'F'},
 #ifdef HAVE_ARITH_CODE
 	{"all-arith",0,&arith_mode,1},
 	{"all-huffman",0,&arith_mode,0},
@@ -193,9 +197,9 @@ METHODDEF(void)	my_error_exit (j_common_ptr cinfo)
 {
 	my_error_ptr myerr = (my_error_ptr)cinfo->err;
 
-	(*cinfo->err->output_message) (cinfo);
+	(*cinfo->err->output_message)(cinfo);
 	if (myerr->jump_set)
-		longjmp(myerr->setjmp_buffer,1);
+		longjmp(myerr->setjmp_buffer, 1);
 	else
 		fatal("fatal error");
 }
@@ -205,20 +209,20 @@ METHODDEF(void) my_output_message (j_common_ptr cinfo)
 {
 	char buffer[JMSG_LENGTH_MAX+1];
 
-	(*cinfo->err->format_message)((j_common_ptr)cinfo,buffer);
-	buffer[sizeof(buffer)-1]=0;
+	(*cinfo->err->format_message)((j_common_ptr)cinfo, buffer);
+	buffer[sizeof(buffer) - 1] = 0;
 
 	if (verbose_mode)
-		fprintf(jpeg_log_fh," (%s) ",buffer);
+		fprintf(jpeg_log_fh, " (%s) ", buffer);
 
 	global_error_counter++;
-	strncpy(last_error,buffer,sizeof(last_error));
+	strncopy(last_error, buffer, sizeof(last_error));
 }
 
 
 void print_usage(void)
 {
-	fprintf(stderr,PROGRAMNAME " v" VERSIO "  " COPYRIGHT "\n");
+	fprintf(stderr,PROGRAMNAME " v" VERSION "  " COPYRIGHT "\n");
 
 	fprintf(stderr,
 		"Usage: " PROGRAMNAME " [options] <filenames> \n\n"
@@ -240,7 +244,7 @@ void print_usage(void)
 		"                    keep old file if the gain is below a threshold (%%)\n"
 #ifdef PARALLEL_PROCESSING
 		"  -w<max>, --workers=<max>\n"
-		"                    set mximum number of parellel threads (default is 1)\n"
+		"                    set maximum number of parallel threads (default is 1)\n"
 #endif
 		"  -b, --csv         print progress info in CSV format\n"
 		"  -o, --overwrite   overwrite target file even if it exists (meaningful\n"
@@ -278,6 +282,8 @@ void print_usage(void)
 #endif
 		"  --stdout          send output to standard output (instead of a file)\n"
 		"  --stdin           read input from standard input (instead of a file)\n"
+		"  --files-stdin     Read names of files to process from stdin\n"
+		"  --files-from=FILE Read names of files to process from a file\n"
 		"  --nofix           skip processing of input files if they contain any errors\n"
 		"\n\n");
 }
@@ -285,29 +291,28 @@ void print_usage(void)
 
 void print_version()
 {
-	struct jpeg_error_mgr jcerr, *err;
-
+	struct jpeg_error_mgr jerr;
 
 #ifdef  __DATE__
-	printf(PROGRAMNAME " v%s  %s (%s)\n",VERSIO,HOST_TYPE,__DATE__);
+	printf(PROGRAMNAME " v%s  %s (%s)\n",VERSION,HOST_TYPE,__DATE__);
 #else
-	printf(PROGRAMNAME " v%s  %s\n",VERSIO,HOST_TYPE);
+	printf(PROGRAMNAME " v%s  %s\n",VERSION,HOST_TYPE);
 #endif
 	printf(COPYRIGHT "\n\n");
 	printf("This program comes with ABSOLUTELY NO WARRANTY. This is free software,\n"
-		"and you are welcome to redistirbute it under certain conditions.\n"
+		"and you are welcome to redistribute it under certain conditions.\n"
 		"See the GNU General Public License for more details.\n\n");
 
-	if (!(err=jpeg_std_error(&jcerr)))
+	if (!jpeg_std_error(&jerr))
 		fatal("jpeg_std_error() failed");
 
 	printf("\nlibjpeg version: %s\n%s\n",
-		err->jpeg_message_table[JMSG_VERSION],
-		err->jpeg_message_table[JMSG_COPYRIGHT]);
+		jerr.jpeg_message_table[JMSG_VERSION],
+		jerr.jpeg_message_table[JMSG_COPYRIGHT]);
 }
 
 
-void parse_arguments(int argc, char **argv, char *dest_path)
+void parse_arguments(int argc, char **argv, char *dest_path, size_t dest_path_len)
 {
 	int opt_index;
 	int i, c;
@@ -319,68 +324,79 @@ void parse_arguments(int argc, char **argv, char *dest_path)
 			break;
 
 		switch (c) {
-		case 'm':
-		{
-			int tmpvar;
 
-			if (sscanf(optarg,"%d",&tmpvar) == 1) {
-				quality=tmpvar;
-				if (quality < 0) quality=0;
-				if (quality > 100) quality=100;
+		case 'm':
+		        {
+				int tmpvar;
+
+				if (sscanf(optarg,"%d",&tmpvar) == 1) {
+					quality=tmpvar;
+					if (quality < 0) quality=0;
+					if (quality > 100) quality=100;
+				}
+				else
+					fatal("invalid argument for -m, --max");
 			}
-			else
-				fatal("invalid argument for -m, --max");
-		}
-		break;
+			break;
+
 		case 'd':
 			if (realpath(optarg,dest_path)==NULL)
 				fatal("invalid destination directory: %s", optarg);
 			if (!is_directory(dest_path))
 				fatal("destination not a directory: %s", dest_path);
-			strncat(dest_path,DIR_SEPARATOR_S,sizeof(dest_path)-strlen(dest_path)-1);
-
+			strncatenate(dest_path, DIR_SEPARATOR_S, dest_path_len);
 			if (verbose_mode)
 				fprintf(stderr,"Destination directory: %s\n",dest_path);
 			dest=1;
 			break;
+
 		case 'v':
 			verbose_mode++;
 			break;
+
 		case 'h':
 			print_usage();
 			exit(0);
 			break;
+
 		case 'q':
 			quiet_mode=1;
 			break;
+
 		case 't':
 			totals_mode=1;
 			break;
+
 		case 'n':
 			noaction=1;
 			break;
+
 		case 'f':
 			force=1;
 			break;
+
 		case 'b':
 			csv=1;
 			quiet_mode=1;
 			break;
-		case '?':
-			exit(1);
+
 		case 'V':
 			print_version();
 			exit(0);
 			break;
+
 		case 'o':
 			overwrite_mode=1;
 			break;
+
 		case 'p':
 			preserve_mode=1;
 			break;
+
 		case 'P':
 			preserve_perms=1;
 			break;
+
 		case 's':
 			save_exif=0;
 			save_iptc=0;
@@ -390,50 +406,70 @@ void parse_arguments(int argc, char **argv, char *dest_path)
 			save_adobe=0;
 			save_jfxx=0;
 			break;
+
 		case 'T':
-		{
-			double tmpvar;
-			if (sscanf(optarg,"%lf", &tmpvar) == 1) {
-				threshold=tmpvar;
-				if (threshold < 0) threshold=0;
-				if (threshold > 100) threshold=100;
-			}
-			else fatal("invalid argument for -T, --threshold");
-		}
-		break;
-		case 'S':
-		{
-			unsigned int tmpvar;
-			if (sscanf(optarg,"%u", &tmpvar) == 1) {
-				if (tmpvar > 0 && tmpvar < 100 &&
-					optarg[strlen(optarg)-1] == '%' ) {
-					target_size=-tmpvar;
-				} else {
-					target_size=tmpvar;
+		        {
+				double tmpvar;
+				if (sscanf(optarg,"%lf", &tmpvar) == 1) {
+					threshold=tmpvar;
+					if (threshold < 0) threshold=0;
+					if (threshold > 100) threshold=100;
 				}
-				quality=100;
+				else fatal("invalid argument for -T, --threshold");
 			}
-			else fatal("invalid argument for -S, --size");
-		}
-		break;
+			break;
+
+		case 'S':
+		        {
+				unsigned int tmpvar;
+				if (sscanf(optarg,"%u", &tmpvar) == 1) {
+					if (tmpvar > 0 && tmpvar < 100 &&
+						optarg[strlen(optarg)-1] == '%' ) {
+						target_size=-tmpvar;
+					} else {
+						target_size=tmpvar;
+					}
+					quality=100;
+				}
+				else fatal("invalid argument for -S, --size");
+			}
+			break;
+
 #ifdef PARALLEL_PROCESSING
 		case 'w':
-		{
-			int tmpvar;
-			if (sscanf(optarg, "%d", &tmpvar) == 1) {
-				if (tmpvar > 0 && tmpvar <= MAX_WORKERS)
-					max_workers = tmpvar;
+		        {
+				int tmpvar;
+				if (sscanf(optarg, "%d", &tmpvar) == 1) {
+					if (tmpvar > 0 && tmpvar <= MAX_WORKERS)
+						max_workers = tmpvar;
+				}
+				else fatal("invalid argument for -w, --workers");
 			}
-			else fatal("invalid argument for -w, --workers");
-		}
-		break;
+			break;
 #endif
+
+		case 'F':
+		        {
+				if (optarg[0] == '-' && optarg[1] == 0) {
+					files_stdin = 1;
+					break;
+				}
+				if (!is_file(optarg, NULL))
+					fatal("argument for --files-from must be a file");
+				if ((files_from = fopen(optarg, "r")) == NULL)
+					fatal("cannot open file: '%s'", optarg);
+			}
+			break;
+
+		case '?':
+			exit(1);
+
 		}
 	}
 
 
 	/* check for '-' option indicating input is from stdin... */
-	i=1;
+	i = optind;
 	while (argv[i]) {
 		if (argv[i][0]=='-' && argv[i][1]==0)
 			stdin_mode=1;
@@ -445,13 +481,19 @@ void parse_arguments(int argc, char **argv, char *dest_path)
 
 	if (all_normal && all_progressive)
 		fatal("cannot specify both --all-normal and --all-progressive");
+
+	if (files_stdin)
+		files_from = stdin;
+
+	if (stdin_mode && files_from == stdin)
+		fatal("cannot specify both --stdin and --files-stdin");
 }
 
 
 void own_signal_handler(int a)
 {
 	if (verbose_mode > 1)
-		fprintf(stderr,PROGRAMNAME ": signal: %d\n",a);
+		fprintf(stderr,PROGRAMNAME ": died from signal: %d\n",a);
 	exit(2);
 }
 
@@ -461,48 +503,38 @@ void write_markers(struct jpeg_decompress_struct *dinfo,
 {
 	jpeg_saved_marker_ptr mrk;
 	int write_marker;
+	const char *s_name;
 
 	if (!cinfo || !dinfo)
 		fatal("invalid call to write_markers()");
 
 	mrk=dinfo->marker_list;
 	while (mrk) {
-		write_marker=0;
+		write_marker = 0;
+		s_name = jpeg_special_marker_name(mrk);
 
-		/* check for markers to save... */
+		/* Check for markers to save... */
 
 		if (save_com && mrk->marker == JPEG_COM)
 			write_marker++;
 
-		if (save_iptc && mrk->marker == IPTC_JPEG_MARKER)
+		if (save_iptc && !strncmp(s_name, "IPTC", 5))
 			write_marker++;
 
-		if (save_exif && mrk->marker == EXIF_JPEG_MARKER &&
-			mrk->data_length >= EXIF_IDENT_STRING_SIZE &&
-			!memcmp(mrk->data, EXIF_IDENT_STRING, EXIF_IDENT_STRING_SIZE))
+		if (save_exif && !strncmp(s_name, "Exif", 5))
 			write_marker++;
 
-		if (save_icc && mrk->marker == ICC_JPEG_MARKER &&
-			mrk->data_length >= ICC_IDENT_STRING_SIZE &&
-			!memcmp(mrk->data, ICC_IDENT_STRING, ICC_IDENT_STRING_SIZE))
+		if (save_icc && !strncmp(s_name, "ICC", 4))
 			write_marker++;
 
-		if (save_xmp && mrk->marker == XMP_JPEG_MARKER &&
-			mrk->data_length >= XMP_IDENT_STRING_SIZE &&
-			!memcmp(mrk->data, XMP_IDENT_STRING, XMP_IDENT_STRING_SIZE))
+		if (save_xmp && !strncmp(s_name, "XMP", 4))
 			write_marker++;
 
-		if (save_jfxx && mrk->marker == JFXX_JPEG_MARKER &&
-			mrk->data_length >= JFXX_IDENT_STRING_SIZE &&
-			!memcmp(mrk->data, JFXX_IDENT_STRING, JFXX_IDENT_STRING_SIZE))
+		if (save_jfxx && !strncmp(s_name, "JFXX", 5))
 			write_marker++;
 
-		if (save_adobe && cinfo->write_Adobe_marker == FALSE &&
-			mrk->marker == ADOBE_JPEG_MARKER &&
-			mrk->data_length >= ADOBE_IDENT_STRING_SIZE &&
-			!memcmp(mrk->data, ADOBE_IDENT_STRING, ADOBE_IDENT_STRING_SIZE)) {
+		if (save_adobe && !strncmp(s_name, "Adobe", 6))
 			write_marker++;
-		}
 
 		if (strip_none)
 			write_marker++;
@@ -510,38 +542,76 @@ void write_markers(struct jpeg_decompress_struct *dinfo,
 
 		/* libjpeg emits some markers automatically so skip these to avoid duplicates... */
 
-		/* skip JFIF (APP0) marker */
-		if ( mrk->marker == JFIF_JPEG_MARKER &&
-			mrk->data_length >= JFIF_IDENT_STRING_SIZE &&
-			!memcmp(mrk->data, JFIF_IDENT_STRING, JFIF_IDENT_STRING_SIZE)) {
-			if (verbose_mode > 2)
-				fprintf(jpeg_log_fh, " (skip JFIF v%u.%02u marker) ",
-					mrk->data[5], mrk->data[6]);
+		if (!strncmp(s_name, "JFIF", 5))
 			write_marker=0;
-		}
 
 
+		if (verbose_mode > 2)
+			fprintf(jpeg_log_fh, " (Marker %s [%s]: %s) ", jpeg_marker_name(mrk->marker),
+				s_name, (write_marker ? "Keep" : "Discard"));
 		if (write_marker)
-			jpeg_write_marker(cinfo,mrk->marker,mrk->data,mrk->data_length);
+			jpeg_write_marker(cinfo, mrk->marker, mrk->data, mrk->data_length);
 
 		mrk=mrk->next;
 	}
 }
 
 
+unsigned int parse_markers(const struct jpeg_decompress_struct *dinfo,
+			char *str, unsigned int str_size, unsigned int *markers_total_size)
+{
+	jpeg_saved_marker_ptr m;
+	unsigned int count = 0;
+	char *seen;
+	size_t marker_types = jpeg_special_marker_types_count();
+	int com_seen = 0;
+	int special;
+
+	if ((seen = malloc(marker_types)) == NULL)
+		fatal("not enough of memory");
+
+	memset(seen, 0, marker_types);
+	str[0] = 0;
+	*markers_total_size = 0;
+
+	m = dinfo->marker_list;
+	while (m) {
+		count++;
+		*markers_total_size += m->data_length;
+
+		if ((special = jpeg_special_marker(m)) >= 0) {
+			if (!seen[special])
+				str_add_list(str, str_size, jpeg_special_marker_types[special].name, ",");
+			seen[special]++;
+		}
+
+		if (m->marker == JPEG_COM && !com_seen) {
+			str_add_list(str, str_size, "COM", ",");
+			com_seen = 1;
+		}
+
+		m = m->next;
+	}
+
+	free(seen);
+
+	return count;
+}
+
+
 int optimize(FILE *log_fh, const char *filename, const char *newname,
 	const char *tmpdir, struct stat *file_stat,
 	double *rate, double *saved)
 {
-	FILE *infile = NULL, *outfile = NULL;
+	FILE *infile = NULL;
+	FILE *outfile = NULL;
 	const char *outfname = NULL;
 	char tmpfilename[MAXPATHLEN];
 	struct jpeg_decompress_struct dinfo;
 	struct jpeg_compress_struct cinfo;
-	struct my_error_mgr jcerr,jderr;
+	struct my_error_mgr jcerr, jderr;
 	JSAMPARRAY buf = NULL;
 
-	jpeg_saved_marker_ptr cmarker;
 	unsigned char *outbuffer = NULL;
 	size_t outbuffersize = 0;
 	unsigned char *inbuffer = NULL;
@@ -550,7 +620,7 @@ int optimize(FILE *log_fh, const char *filename, const char *newname,
 
 	jvirt_barray_ptr *coef_arrays = NULL;
 	char marker_str[256];
-	int marker_in_count, marker_in_size;
+	unsigned int marker_in_count, marker_in_size;
 
 	long insize = 0, outsize = 0, lastsize = 0;
 	long inpos;
@@ -561,14 +631,14 @@ int optimize(FILE *log_fh, const char *filename, const char *newname,
 
 	jpeg_log_fh = log_fh;
 
-	/* initialize decompression object */
+	/* Initialize decompression object */
 	dinfo.err = jpeg_std_error(&jderr.pub);
 	jpeg_create_decompress(&dinfo);
 	jderr.pub.error_exit=my_error_exit;
 	jderr.pub.output_message=my_output_message;
 	jderr.jump_set = 0;
 
-	/* initialize compression object */
+	/* Initialize compression object */
 	cinfo.err = jpeg_std_error(&jcerr.pub);
 	jpeg_create_compress(&cinfo);
 	jcerr.pub.error_exit=my_error_exit;
@@ -593,7 +663,7 @@ retry_point:
 	}
 
 	if (setjmp(jderr.setjmp_buffer)) {
-		/* error handler for decompress */
+		/* Error handler for decompress */
 	abort_decompress:
 		jpeg_abort_decompress(&dinfo);
 		fclose(infile);
@@ -613,7 +683,7 @@ retry_point:
 		fflush(log_fh);
 	}
 
-	/* prepare to decompress */
+	/* Prepare to decompress */
 	if (stdin_mode || stdout_mode) {
 		if (inbuffer)
 			free(inbuffer);
@@ -630,69 +700,33 @@ retry_point:
 	jpeg_custom_src(&dinfo, infile, &inbuffer, &inbuffersize, &inbufferused, 65536);
 	jpeg_read_header(&dinfo, TRUE);
 
-	/* check for Exif/IPTC/ICC/XMP markers */
-	marker_str[0]=0;
-	marker_in_count=0;
-	marker_in_size=0;
-	cmarker=dinfo.marker_list;
-
-	while (cmarker) {
-		marker_in_count++;
-		marker_in_size+=cmarker->data_length;
-
-		if (cmarker->marker == EXIF_JPEG_MARKER &&
-			cmarker->data_length >= EXIF_IDENT_STRING_SIZE &&
-			!memcmp(cmarker->data,EXIF_IDENT_STRING,EXIF_IDENT_STRING_SIZE))
-			strncat(marker_str,"Exif ",sizeof(marker_str)-strlen(marker_str)-1);
-
-		if (cmarker->marker == IPTC_JPEG_MARKER)
-			strncat(marker_str,"IPTC ",sizeof(marker_str)-strlen(marker_str)-1);
-
-		if (cmarker->marker == ICC_JPEG_MARKER &&
-			cmarker->data_length >= ICC_IDENT_STRING_SIZE &&
-			!memcmp(cmarker->data,ICC_IDENT_STRING,ICC_IDENT_STRING_SIZE))
-			strncat(marker_str,"ICC ",sizeof(marker_str)-strlen(marker_str)-1);
-
-		if (cmarker->marker == XMP_JPEG_MARKER &&
-			cmarker->data_length >= XMP_IDENT_STRING_SIZE &&
-			!memcmp(cmarker->data,XMP_IDENT_STRING,XMP_IDENT_STRING_SIZE))
-			strncat(marker_str,"XMP ",sizeof(marker_str)-strlen(marker_str)-1);
-
-		if (cmarker->marker == JFXX_JPEG_MARKER &&
-			cmarker->data_length >= JFXX_IDENT_STRING_SIZE &&
-			!memcmp(cmarker->data, JFXX_IDENT_STRING, JFXX_IDENT_STRING_SIZE))
-			strncat(marker_str,"JFXX ",sizeof(marker_str)-strlen(marker_str)-1);
-
-		cmarker=cmarker->next;
-	}
-
-
+	/* Check for known (Exif, IPTC, ICC , XMP, ...) markers */
+	marker_in_count = parse_markers(&dinfo, marker_str, sizeof(marker_str),
+					&marker_in_size);
 	if (verbose_mode > 1) {
 		fprintf(log_fh,"%d markers found in input file (total size %d bytes)\n",
 			marker_in_count,marker_in_size);
 		fprintf(log_fh,"coding: %s\n", (dinfo.arith_code == TRUE ? "Arithmetic" : "Huffman"));
 	}
+
+
 	if (!retry && (!quiet_mode || csv)) {
 		fprintf(log_fh,csv ? "%dx%d,%dbit,%c," : "%dx%d %dbit %c ",(int)dinfo.image_width,
 			(int)dinfo.image_height,(int)dinfo.num_components*8,
 			(dinfo.progressive_mode?'P':'N'));
-
-		if (!csv) {
+		if (!csv)
 			fprintf(log_fh,"%s",marker_str);
-			if (dinfo.saw_Adobe_marker) fprintf(log_fh,"Adobe ");
-			if (dinfo.saw_JFIF_marker) fprintf(log_fh,"JFIF ");
-		}
 		fflush(log_fh);
 	}
 
 	if ((insize=file_size(infile)) < 0)
 		fatal("failed to stat() input file");
 
-	/* decompress the file */
+	/* Decompress the image */
 	if (quality >= 0 && !retry) {
 		jpeg_start_decompress(&dinfo);
 
-		/* allocate line buffer to store the decompressed image */
+		/* Allocate line buffer to store the decompressed image */
 		buf = malloc(sizeof(JSAMPROW)*dinfo.output_height);
 		if (!buf) fatal("not enough memory");
 		for (j=0;j<dinfo.output_height;j++) {
@@ -715,10 +749,11 @@ retry_point:
 		}
 	}
 
-	inpos=ftell(infile);
+	inpos = ftell(infile);
 	if (inpos > 0 && inpos < insize) {
 		if (!quiet_mode)
-			fprintf(log_fh, " (Extraneous data found after end of JPEG image) ");
+			fprintf(log_fh, " (%lu bytes extraneous data found after end of image) ",
+				insize - inpos);
 		if (nofix_mode)
 			global_error_counter++;
 	}
@@ -732,7 +767,7 @@ retry_point:
 		insize = inbufferused;
 
 	if (nofix_mode && global_error_counter != 0) {
-		/* skip files containing any errors (warnings) */
+		/* Skip files containing any errors (or warnings) */
 		goto abort_decompress;
 	}
 
@@ -746,7 +781,7 @@ retry_point:
 
 
 	if (setjmp(jcerr.setjmp_buffer)) {
-		/* error handler for compress failures */
+		/* Error handler for compress failures */
 		jpeg_abort_compress(&cinfo);
 		jpeg_abort_decompress(&dinfo);
 		fclose(infile);
@@ -766,14 +801,14 @@ retry_point:
 	searchdone = 0;
 	oldquality = 200;
 	if (target_size != 0) {
-		/* always start with quality 100 if -S option specified... */
+		/* Always start with quality 100 if -S option specified... */
 		quality = 100;
 	}
 
 
 binary_search_loop:
 
-	/* allocate memory buffer that should be large enough to store the output JPEG... */
+	/* Allocate memory buffer that should be large enough to store the output JPEG... */
 	if (outbuffer)
 		free(outbuffer);
 	outbuffersize=insize + 32768;
@@ -786,7 +821,7 @@ binary_search_loop:
 
 
 	if (quality >= 0 && !retry) {
-		/* lossy "optimization" ... */
+		/* Lossy "optimization" ... */
 
 		cinfo.in_color_space=dinfo.out_color_space;
 		cinfo.input_components=dinfo.output_components;
@@ -798,11 +833,12 @@ binary_search_loop:
 		if (jpeg_c_int_param_supported(&cinfo, JINT_DC_SCAN_OPT_MODE))
 			jpeg_c_set_int_param(&cinfo, JINT_DC_SCAN_OPT_MODE, 1);
 #endif
-		if (all_normal) {
-			/* Explicitly disables progressive if libjpeg had it on by default */
+		if (all_normal || (!dinfo.progressive_mode && !all_progressive)) {
+			/* Explicitly disable progressive mode. */
 			cinfo.scan_info = NULL;
 			cinfo.num_scans = 0;
-		} else if (dinfo.progressive_mode || all_progressive) {
+		} else if (all_progressive || dinfo.progressive_mode) {
+			/* Enable progressive mode. */
 			jpeg_simple_progression(&cinfo);
 		}
 		cinfo.optimize_coding = TRUE;
@@ -818,28 +854,29 @@ binary_search_loop:
 		j=0;
 		jpeg_start_compress(&cinfo,TRUE);
 
-		/* write markers */
+		/* Write markers */
 		write_markers(&dinfo,&cinfo);
 
-		/* write image */
+		/* Write image */
 		while (cinfo.next_scanline < cinfo.image_height) {
 			jpeg_write_scanlines(&cinfo,&buf[cinfo.next_scanline],
 					dinfo.output_height);
 		}
 
 	} else {
-		/* lossless "optimization" ... */
+		/* Lossless optimization ... */
 
 		jpeg_copy_critical_parameters(&dinfo, &cinfo);
 #ifdef HAVE_JINT_DC_SCAN_OPT_MODE
 		if (jpeg_c_int_param_supported(&cinfo, JINT_DC_SCAN_OPT_MODE))
 			jpeg_c_set_int_param(&cinfo, JINT_DC_SCAN_OPT_MODE, 1);
 #endif
-		if (all_normal) {
-			/* Explicitly disables progressive if libjpeg had it on by default */
+		if (all_normal || (!dinfo.progressive_mode && !all_progressive)) {
+			/* Explicitly disable progressive mode. */
 			cinfo.scan_info = NULL;
 			cinfo.num_scans = 0;
-		} else if ( dinfo.progressive_mode || all_progressive ) {
+		} else if (all_progressive || dinfo.progressive_mode) {
+			/* Enable progressive mode. */
 			jpeg_simple_progression(&cinfo);
 		}
 		cinfo.optimize_coding = TRUE;
@@ -852,10 +889,10 @@ binary_search_loop:
 			cinfo.write_JFIF_header = FALSE;
 		}
 
-		/* write image */
+		/* Write image */
 		jpeg_write_coefficients(&cinfo, coef_arrays);
 
-		/* write markers */
+		/* Write markers */
 		write_markers(&dinfo,&cinfo);
 
 	}
@@ -864,7 +901,7 @@ binary_search_loop:
 	outsize=outbuffersize;
 
 	if (target_size != 0 && !retry) {
-		/* perform (binary) search to try to reach target file size... */
+		/* Perform (binary) search to try to reach target file size... */
 
 		long osize = outsize/1024;
 		long isize = insize/1024;
@@ -1055,7 +1092,7 @@ binary_search_loop:
 			fprintf(log_fh,csv ? "skipped\n" : "skipped.\n");
 		if (stdout_mode) {
 			set_filemode_binary(stdout);
-			if (fwrite(inbuffer,insize,1,stdout) != 1)
+			if (fwrite(inbuffer, inbufferused, 1, stdout) != 1)
 				fatal("%s, write failed to stdout",
 					(stdin_mode ? "stdin" : filename));
 		}
@@ -1101,7 +1138,7 @@ int wait_for_worker(FILE *log_fh)
 		}
 	}
 	if (!w)
-		fatal("Unknow worker[%d] process found\n", pid);
+		fatal("Unknown worker[%d] process found\n", pid);
 
 	if (WIFEXITED(wstatus)) {
 		e = WEXITSTATUS(wstatus);
@@ -1166,8 +1203,9 @@ int wait_for_worker(FILE *log_fh)
 int main(int argc, char **argv)
 {
 	struct stat file_stat;
-	char tmpfilename[MAXPATHLEN],tmpdir[MAXPATHLEN];
-	char newname[MAXPATHLEN], dest_path[MAXPATHLEN];
+	char tmpfilename[MAXPATHLEN + 1],tmpdir[MAXPATHLEN + 1];
+	char newname[MAXPATHLEN + 1], dest_path[MAXPATHLEN + 1];
+	char namebuf[MAXPATHLEN + 2];
 	const char *filename;
 	volatile int i;
 	int res;
@@ -1179,7 +1217,7 @@ int main(int argc, char **argv)
 	pid_t pid;
 	int j;
 
-	/* allocate table to keep track of child processes... */
+	/* Allocate table to keep track of child processes... */
 	if (!(workers = malloc(sizeof(struct worker) * MAX_WORKERS)))
 		fatal("not enough memory");
 	for (i = 0; i < MAX_WORKERS; i++) {
@@ -1192,8 +1230,8 @@ int main(int argc, char **argv)
 	signal(SIGINT,own_signal_handler);
 	signal(SIGTERM,own_signal_handler);
 
-	/* parse command line parameters */
-	parse_arguments(argc, argv, dest_path);
+	/* Parse command line parameters */
+	parse_arguments(argc, argv, dest_path, sizeof(dest_path));
 	log_fh = (stdout_mode ? stderr : stdout);
 
 	if (verbose_mode) {
@@ -1219,13 +1257,13 @@ int main(int argc, char **argv)
 
 
 	if (stdin_mode) {
-		/* process just one file, if source is stdin... */
+		/* Process just one file, if source is stdin... */
 		res = optimize(stderr, NULL, NULL, NULL, &file_stat, NULL, NULL);
 		return (res == 0 ? 0 : 1);
 	}
 
 	i=(optind > 0 ? optind : 1);
-	if (argc <= i) {
+	if (files_from == NULL && argc <= i) {
 		if (!quiet_mode)
 			fprintf(stderr, PROGRAMNAME ": file argument(s) missing\n"
 				"Try '" PROGRAMNAME " --help' for more information.\n");
@@ -1233,15 +1271,21 @@ int main(int argc, char **argv)
 	}
 
 
-	/* loop to process the input files */
+	/* Main loop to process input files */
 	do {
-		filename = argv[i];
+		if (files_from) {
+			if (!fgetstr(namebuf, sizeof(namebuf), files_from))
+				break;
+			filename = namebuf;
+		} else {
+			filename = argv[i];
+		}
 
-		if (i >= argc || filename[0] == 0)
+		if (*filename == 0)
 			continue;
 		if (verbose_mode > 1)
 			fprintf(log_fh, "processing file: %s\n", filename);
-		if (strlen(filename) >= MAXPATHLEN) {
+		if (strnlen(filename, MAXPATHLEN + 1) > MAXPATHLEN) {
 			warn("skipping too long filename: %s", filename);
 			continue;
 		}
@@ -1249,15 +1293,15 @@ int main(int argc, char **argv)
 		if (!noaction) {
 			/* generate tmp dir & new filename */
 			if (dest) {
-				STRNCPY(tmpdir, dest_path, sizeof(tmpdir));
-				STRNCPY(newname, dest_path, sizeof(newname));
+				strncopy(tmpdir, dest_path, sizeof(tmpdir));
+				strncopy(newname, dest_path, sizeof(newname));
 				if (!splitname(filename, tmpfilename, sizeof(tmpfilename)))
 					fatal("splitname() failed for: %s", filename);
-				strncat(newname, tmpfilename, sizeof(newname)-strlen(newname)-1);
+				strncatenate(newname, tmpfilename, sizeof(newname));
 			} else {
 				if (!splitdir(filename, tmpdir, sizeof(tmpdir)))
 					fatal("splitdir() failed for: %s", filename);
-				STRNCPY(newname, filename, sizeof(newname));
+				strncopy(newname, filename, sizeof(newname));
 			}
 		}
 
@@ -1288,7 +1332,9 @@ int main(int argc, char **argv)
 			if (pid < 0)
 				fatal("fork() failed");
 			if (pid == 0) {
-				/* child process starts here... */
+				/* Child process starts here... */
+				if (files_from)
+					fclose(files_from);
 				close(pipe_fd[0]);
 				FILE *p;
 
@@ -1300,7 +1346,7 @@ int main(int argc, char **argv)
 					fprintf(p, "\n\nSTATS\n%lf\n%lf\n", rate, saved);
 				exit(res);
 			} else {
-				/* parent continues here... */
+				/* Parent continues here... */
 				close(pipe_fd[1]);
 
 				w = NULL;
@@ -1336,7 +1382,7 @@ int main(int argc, char **argv)
 			}
 		}
 
-	} while (++i < argc);
+	} while (files_from || ++i < argc);
 
 
 #ifdef PARALLEL_PROCESSING
